## Graph neural networks
A graph neural network is a neural network for processing data that can be represented as a graph.
In mva-trainer graph neural networks are realised using the [PyTorch Geometric library](https://pytorch-geometric.readthedocs.io/en/latest/).
The building of graphs happens automatically in the framework.
Just like for DNNs it is based on information passed via the config files.
Depending on the type of the graph and the corresponding network the necessary options differ slightly.

## Homogeneous and heterogeneous graphs

In the following the two main types of graphs, homogeneous and heterogeneous, and how they are created in mva-trainer, are highlighted.
Afterwards the structure of GNN models as implemented in mva-trainer is explained.
In general, graphs can be homogeneous or heterogeneous.
They represent networks with nodes and edges connecting the individual nodes.
Fundamentally homogeneous and heterogeneous graphs differ in how these nodes and edges are defined.

## Homogeneous graphs

A homogeneous graph (in machine learning) consists of nodes (black dots in the figure) and edges of identical dimensions (solid blue lines).
This means that each node in the graph has precisely `F` features, and each edge has exactly `E` edge features.
All connections are represented via a square adjacency matrix which determines whether there is a connection (edge) between any two nodes of the graph:

![Simple graph](../img/Graph_01.png)

The power of a GNN is that its input (i.e.) the graph can technically have any size.
This means that the graph could have 3, 5, 10 or 100 nodes, and the GNN would still be able to work with it.
In HEP, the nodes often represent (e.g.) reconstructed objects in the detector.
For this example, we will assume that these are jets and their reconstructed properties.
Naturally, every jet has the same properties; hence, the feature vector `F` for every jet has the same dimension.
Furthermore, we can assign edge attributes; in the example, the distance between these jets in the detector is measured in eta and phi.
These would then be the edge attributes.

Technically this is done in the following way (here with 3 jets):
```
GRAPHNODE
        Name = "Jet_1"
        Type = "Jet"
        Features = "Jet_1_pT", "Jet_1_eta", "Jet_1_phi"
        Targets = "Jet_1", "Jet_3"
        EdgeFeatures = "Jet_1_2_deta","Jet_1_2_dphi"|"Jet_1_3_deta","Jet_1_3_dphi"
	
GRAPHNODE
        Name = "Jet_2"
        Type = "Jet"
        Features = "Jet_2_pT", "Jet_2_eta", "Jet_2_phi"
        Targets = "Jet_1", "Jet_3"
        EdgeFeatures = "Jet_1_2_deta","Jet_2_3_dphi"|"Jet_2_3_deta","Jet_2_3_dphi"

GRAPHNODE
        Name = "Jet_3"
        Type = "Jet"
        Features = "Jet_3_pT", "Jet_3_eta", "Jet_3_phi"
        Targets = "Jet_1", "Jet_2"
        EdgeFeatures = "Jet_1_3_deta","Jet_1_3_dphi"|"Jet_2_3_deta","Jet_2_3_dphi"
```

Each `GRAPHNODE` object defines one new node.
The `type` will be necessary once we start talking about *heterogeneous* jets.
The `Features` represent the node's features (`F`), and the targets are all the other `GRAPHNODE`'s this node connects to.
Finally the `EdgeFeatures` is a list of features per edge in the same order as the targets, separated by `|`. I.e. the first list represents the edge features connecting the particular `GraphNode` with its first target.

What if there are fewer objects than I defined in the config file?
In general, the way graphs should be defined in this framework is by defining the maximum possible graph.
I.e. if you want to train up to 10 jets in the example above, you should define 10 `GRAPHNODE` objects with the respective connections.
If an event has less than 10 jets, you will need to use the `PruneIfValue` option in the `GRAPHNODE` block and pass a list of values with the same length as `F`.
The code will then check if any of the values in the feature vector of this node has this value (e.g. if you assign -1 to values which are not filled) and subsequently prune the node and remove its edges (white node and dashed blue lines):

![Missing Node](../img/Graph_missing.png)

This way you do not have to worry about the size of the graph, and let us face it, who trusts 10 jet events?

## Heterogeneous graphs

Heterogeneous graphs are a more general form of homogeneous graphs in the sense that there can be different types of nodes and edges, they no longer need to have the same feature and edge dimensions.
Instead the graph is "split" into nodes and edges of different types:

![Missing Node](../img/Graph_02.png)

In this example we look at two different types of nodes (black and green)
These could now be jets and leptons with different properties.
Consequently there are now three different types of connections:

* lepton - jet
* lepton - lepton
* jet - jet

All of these edges can also have different properties.
Technically this is realised via the `type` option in each `GRAPHNODE` block.
Via this option the type of each node is defined and identical types are grouped.
Hence any form of event for which one could draw a feynman diagram can be defined.
Again the pruning mechanism works in the same way as in the homogeneous example.

## Global graph features

Apart from the node and edge features graphs might also include global features.
In the example above these would be parameters that describe the entire event.
To incorporate these types of variables into the graph a dedicated `GRAPHNODE` block can be used:

```
GRAPHNODE
        Name = "Global Features"                                                                                                                                                                          
        Type = "Global"
	Features = "variable_1","variable_2","variable_3"
```
Here the reserved `Global` option is used as a `type`.
Nodes which are defined as global are used to build separate densely connected networks.
Their architecture is steered via the `GlobalNodes` and `GlobalActivationFunctions` options.
Their output is then concatenated together with the output of the graph convolutional layers.

## Structure of a full GNN in mva-trainer

Graph neural networks (GNNs) in mva-trainer consist of three parts:

* a graph convolution model
* a global model
* a fully connected model

The graph convolution model takes the graph data and applies user defined convulitions.
The global model takes global data and feeds it through a dense neural network.
The fully connected model (here also called final model) then takes the interim outputs of these two models and combines them.
All three are combined into one GNN model which is then trained:

![Simple graph](../img/GNN_Model.png)